# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.

import importlib.metadata

import tuzue.inspector
import tuzue.ui.tcurses
import tuzue.view


def version():
    return importlib.metadata.version("tuzue")


def navigate(struct, title=""):
    view = tuzue.view.View(items=struct, title=title)
    done = None
    with tuzue.ui.tcurses.context() as ui:
        while not done:
            ui.show(view)
            done = ui.interact(view)
    return view.selected_item()


def inspect(*args, **kwargs):
    return tuzue.inspector.inspect(*args, **kwargs)



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.
"""
"inspect" function that is specially useful in PDB
"""

import collections.abc as abc

import tuzue


def process(objdict, key, value, f=repr, pre="[", pos="]"):
    item = "%s%s%s = %s" % (pre, f(key), pos, repr(value))
    path = "%s%s%s" % (pre, f(key), pos)
    objdict[item] = (path, value)
    return item


def identity(x):
    return x


def generator(obj, objdict):
    yield "."
    yield ".."
    if isinstance(obj, abc.Mapping):
        for key, value in obj.items():
            yield process(objdict, key, value)
    elif isinstance(obj, abc.Sequence):
        for key, value in enumerate(obj):
            yield process(objdict, key, value)
    for key in dir(obj):
        if not key.startswith("_"):
            try:
                value = getattr(obj, key)
            except AttributeError as e:
                value = e
            yield process(objdict, key, value, identity, ".", "")
    for key in dir(obj):
        if key.startswith("_"):
            try:
                value = getattr(obj, key)
            except AttributeError as e:
                value = e
            yield process(objdict, key, value, identity, ".", "")


class Inspector:
    def __init__(self, ui):
        self.ui = ui
        self.done = False
        self.result = None

    def inspect(self, path0, obj, lvl=0):
        objdict = {}
        view = tuzue.view.View(title="".join(path0), generator=generator(obj, objdict))
        while not self.done:
            done = False
            while not done:
                self.ui.show(view)
                done = self.ui.interact(view)
            obj = view.selected_item()
            if obj == ".":
                self.done = True
                return
            if obj == "..":
                if lvl == 0:
                    self.done = True
                return
            objdata = objdict[obj]
            path = list(path0)
            path.append(objdata[0])
            self.result = (path, objdata[1])
            self.inspect(path, objdata[1], lvl + 1)


def inspect(obj=None, name=None):
    if obj is None:
        obj = globals()
    result = None
    path = []
    if name:
        path = [name]
    with tuzue.ui.tcurses.context() as ui:
        inspector = Inspector(ui)
        result = None
        while not inspector.done:
            inspector.inspect(path, obj)
        result = inspector.result
    print("".join(result[0]), "=", repr(result[1]))
    return result



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.
"""
A View object has the state of a particular view. It's not
necessarily the current view, but it can be so by using the Screen
instance.
"""

import tuzue.binput


class View:
    def __init__(self, title="", items=None, generator=None):
        # One of the mutually-exclusive arguments must be provided:
        assert (items is None) != (generator is None)
        # All items:
        self.items_all = items or []
        # item_generator, when in use:
        self.item_generator = generator
        # Effective items, filtered by input:
        self.items = []
        # Selected item, identified by items index:
        self.selected_idx = None
        # Screen height, or number of visible items:
        self.screen_height = None
        # Screen starts at this self.item idx:
        self.screen_idx = None
        # Self-managed input object:
        self.binput = tuzue.binput.Binput()
        # Title, shown in header:
        self.title = title
        # Reset to sync selected_idx with items:
        self.reset()

    def reset(self):
        """
        Reset self.items to self.items_all, as if we had no filter input.
        Reset selected_idx.
        """
        self.items = list(self.items_all) or []
        self.selected_idx = 0 if self.items else None
        self.screen_idx = 0 if self.screen_height else None

    # Item generation methods:

    def item_generate(self):
        """
        Generate one item using self.item_generator, if possible, and append it to
        self.items_all.

        Returns True if an item was generated, or False otherwise - if we are already
        done, for instance.
        """
        if not self.item_generator:
            return False
        try:
            wasempty = self.items == []
            item = next(self.item_generator)
            self.items_all.append(item)
            if self.item_filter(item):
                self.items.append(item)
                if self.selected_idx is None and wasempty:
                    self.selected_idx = 0
            return True
        except StopIteration:
            self.item_generator = None
            return False

    def items_generate_all(self):
        while self.item_generate():
            pass

    # Item filtering methods:

    def item_filter(self, item):
        """Returns True if the provided item should be shown, given the current input"""
        return not self.binput.string or self.binput.string in item

    def items_update(self):
        """Resets and updates the whole self.items list using the current input;
        also resets selected_idx if necessary"""
        selected_item = None
        if self.selected_idx is not None:
            selected_item = self.items[self.selected_idx]
        self.items = []
        self.selected_idx = None
        idx = 0
        for item in self.items_all:
            if self.item_filter(item):
                self.items.append(item)
                if item == selected_item:
                    self.selected_idx = idx
                idx += 1
        if self.selected_idx is None and self.items:
            self.selected_idx = 0
        if not self.selected_in_screen():
            self.screen_center()

    # Selected idx/items methods:

    def selected_idx_set(self, idx):
        if idx < 0 or idx >= len(self.items):
            return
        self.selected_idx = idx

    def selected_item(self):
        if self.selected_idx is not None:
            return self.items[self.selected_idx]
        return None

    def selected_in_screen(self):
        if self.screen_height is None or self.selected_idx is None:
            return True
        return (
            self.screen_idx <= self.selected_idx
            and self.selected_idx < self.screen_idx + self.screen_height
        )

    # Screen methods, used to generate the concrete view:

    def screen_height_set(self, height):
        old_screen_height = self.screen_height
        self.screen_height = height
        if height != old_screen_height:
            self.reset()

    def screen_items(self):
        screen_idx = self.screen_idx or 0
        for line, item in enumerate(self.items[screen_idx:]):
            if self.screen_height is not None and line >= self.screen_height:
                break
            yield item

    def screen_selected_line(self):
        if self.selected_idx is None:
            return None
        return self.selected_idx - (self.screen_idx or 0)

    def screen_center(self):
        mid = self.screen_height // 2
        self.screen_idx = max(0, self.selected_idx - mid)
        # If we can scroll down to show more items:
        items_len = len(self.items)
        if (
            items_len > self.screen_height
            and self.selected_idx > items_len - self.screen_height
        ):
            self.screen_idx = items_len - self.screen_height

    # Key reactors:

    def typed(self, char):
        self.binput.typed(char)
        self.items_update()

    def key_enter(self, key=None, keyname=None):
        return True

    def key_down(self, key=None, keyname=None):
        self.selected_idx_set(self.selected_idx + 1)
        if not self.selected_in_screen():
            self.screen_idx += 1

    def key_up(self, key=None, keyname=None):
        self.selected_idx_set(self.selected_idx - 1)
        if not self.selected_in_screen():
            self.screen_idx -= 1

    def key_pgup(self, key=None, keyname=None):
        self.screen_idx = self.screen_idx - self.screen_height + 1
        if self.screen_idx < 0:
            self.screen_idx = 0
            self.selected_idx_set(0)
        else:
            self.selected_idx_set(self.screen_idx + self.screen_height - 1)

    def key_pgdown(self, key=None, keyname=None):
        self.screen_idx = self.screen_idx + self.screen_height - 1
        if self.screen_idx > len(self.items) - self.screen_height:
            self.screen_idx = len(self.items) - self.screen_height
            self.selected_idx_set(len(self.items) - 1)
        else:
            self.selected_idx_set(self.screen_idx)

    def key_home(self, key=None, keyname=None):
        self.screen_idx = 0
        self.selected_idx_set(0)

    def key_end(self, key=None, keyname=None):
        self.screen_idx = len(self.items) - self.screen_height
        self.selected_idx_set(len(self.items) - 1)

    def key_backspace(self, key=None, keyname=None):
        self.binput.key_backspace()
        self.items_update()

    def key_delete(self, key=None, keyname=None):
        self.binput.key_delete()
        self.items_update()

    def key_left(self, key=None, keyname=None):
        self.binput.key_left()

    def key_right(self, key=None, keyname=None):
        self.binput.key_right()

    def key_bol(self, key=None, keyname=None):
        self.binput.key_bol()

    def key_eol(self, key=None, keyname=None):
        self.binput.key_eol()

    def key_killbol(self, key=None, keyname=None):
        self.binput.key_killbol()

    def key_killeol(self, key=None, keyname=None):
        self.binput.key_killeol()

    def key_killwordleft(self, key=None, keyname=None):
        self.binput.key_killwordleft()
        self.items_update()



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.

import unittest

import tuzue.view


class TestView(unittest.TestCase):
    def test_empty(self):
        view = tuzue.view.View(items=[])
        self.assertEqual(list(view.screen_items()), [])
        self.assertEqual(view.selected_item(), None)

    def test_generator(self):
        itemlist = [str(i) for i in range(0, 20)]
        generator = (i for i in itemlist)
        view = tuzue.view.View(generator=generator)
        view.items_generate_all()
        self.assertEqual(list(view.screen_items()), itemlist)
        view.items_update()
        self.assertEqual(list(view.screen_items()), itemlist)

    def test_filter(self):
        itemlist = [str(i) for i in range(0, 20)]
        view = tuzue.view.View(items=itemlist)
        self.assertEqual(view.selected_item(), "0")
        view.typed("1")
        self.assertEqual(
            list(view.screen_items()),
            ["1"] + ["1%d" % i for i in range(0, 10)],
        )
        self.assertEqual(view.selected_item(), "1")
        view.key_up()
        self.assertEqual(view.selected_item(), "1")
        view.key_down()
        self.assertEqual(view.selected_item(), "10")
        view.typed("0")
        self.assertEqual(view.selected_item(), "10")
        view.key_backspace()
        self.assertEqual(view.selected_item(), "10")
        view.key_up()
        self.assertEqual(view.selected_item(), "1")
        view.typed("z")
        self.assertEqual(list(view.screen_items()), [])
        self.assertEqual(view.selected_item(), None)

    def test_screen(self):
        itemlist = [str(i) for i in range(0, 10)]
        view = tuzue.view.View(items=itemlist)
        view.screen_height_set(3)
        self.assertEqual(list(view.screen_items()), ["0", "1", "2"])
        for i in range(0, 3):
            self.assertEqual(view.selected_idx, i)
            self.assertEqual(view.selected_item(), str(i))
            self.assertEqual(view.screen_selected_line(), i)
            view.key_down()
        for i in range(3, 10):
            self.assertEqual(view.selected_idx, i)
            self.assertEqual(view.selected_item(), str(i))
            self.assertEqual(view.screen_selected_line(), 2)
            view.key_down()
        for i in range(9, 6, -1):
            self.assertEqual(view.selected_idx, i)
            self.assertEqual(view.selected_item(), str(i))
            self.assertEqual(view.screen_selected_line(), i - 7)
            view.key_up()
        for i in range(6, -1, -1):
            self.assertEqual(view.selected_idx, i)
            self.assertEqual(view.selected_item(), str(i))
            self.assertEqual(view.screen_selected_line(), 0)
            view.key_up()
        self.assertEqual(view.selected_idx, 0)
        self.assertEqual(view.selected_item(), "0")
        self.assertEqual(view.screen_selected_line(), 0)

    def test_screen_filter(self):
        itemlist = [str(i) for i in range(0, 20)]
        view = tuzue.view.View(items=itemlist)
        view.screen_height_set(3)
        self.assertEqual(list(view.screen_items()), ["0", "1", "2"])
        self.assertEqual(view.selected_item(), "0")
        view.typed("1")
        self.assertEqual(list(view.screen_items()), ["1", "10", "11"])
        self.assertEqual(view.selected_idx, 0)
        self.assertEqual(view.selected_item(), "1")
        self.assertEqual(view.screen_selected_line(), 0)
        view.key_down()
        self.assertEqual(list(view.screen_items()), ["1", "10", "11"])
        self.assertEqual(view.selected_idx, 1)
        self.assertEqual(view.selected_item(), "10")
        self.assertEqual(view.screen_selected_line(), 1)
        view.typed("9")
        self.assertEqual(list(view.screen_items()), ["19"])
        self.assertEqual(view.selected_item(), "19")
        self.assertEqual(view.screen_selected_line(), 0)
        view.key_backspace()
        self.assertEqual(list(view.screen_items()), ["17", "18", "19"])
        self.assertEqual(view.selected_item(), "19")
        self.assertEqual(view.screen_selected_line(), 2)

    def test_screen_pg(self):
        itemlist = [str(i) for i in range(0, 6)]
        view = tuzue.view.View(items=itemlist)
        view.screen_height_set(3)
        view.key_pgdown()
        self.assertEqual(list(view.screen_items()), ["2", "3", "4"])
        self.assertEqual(view.selected_item(), "2")
        view.key_pgup()
        self.assertEqual(list(view.screen_items()), ["0", "1", "2"])
        self.assertEqual(view.selected_item(), "2")
        view.key_pgdown()
        view.key_pgdown()
        self.assertEqual(list(view.screen_items()), ["3", "4", "5"])
        self.assertEqual(view.selected_item(), "5")
        view.key_pgup()
        view.key_pgup()
        view.key_pgup()
        self.assertEqual(list(view.screen_items()), ["0", "1", "2"])
        self.assertEqual(view.selected_item(), "0")

    def test_screen_homeend(self):
        itemlist = [str(i) for i in range(0, 6)]
        view = tuzue.view.View(items=itemlist)
        view.screen_height_set(3)
        view.key_end()
        self.assertEqual(list(view.screen_items()), ["3", "4", "5"])
        self.assertEqual(view.selected_item(), "5")
        view.key_home()
        self.assertEqual(list(view.screen_items()), ["0", "1", "2"])
        self.assertEqual(view.selected_item(), "0")



# Copyright (C) 2023 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.
"""
Logger module for tuzue that facilitates logging configuration changes
"""

import logging

logger = logging.getLogger("tuzue")


def debug_on(filename):
    handler = logging.FileHandler(filename, mode="w")
    formatter = logging.Formatter("%(asctime)s %(message)s")
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    global logger
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.
"""
UiCurses class that represents the full state of the curses screen
itself and is able to sync it with the provided view.

Should be instantiated as a singleton.

This class is used to display a particular View instance, and allows
us to have several views on the stack and switch between then
efficiently. It serves as a bridge between the View class and the
downstream ui implementation - but for now only curses is available.
"""


import curses
from contextlib import contextmanager
from typing import Dict

from tuzue.logger import logger
from tuzue.view import View


class CursesError(Exception):
    pass


class CursesWin:
    """Class that encapsulates a curses window"""

    def __init__(self, label, height, width, line, col):
        self.height = height
        self.width = width
        self.line = line
        self.col = col
        self.label = label
        self.win = curses.newwin(height, width, line, col)

    def erase(self):
        self.win.erase()

    def noutrefresh(self):
        self.win.noutrefresh()

    def addstr(self, line, col, string, *args):
        available = self.width - col
        if len(string) > available:
            string = string[0 : (self.width - col - 4)] + "..."
        try:
            self.win.addstr(line, col, string, *args)
        except curses.error:
            raise CursesError("win {} could not addstr {}\n".format(self.label, string))

    def set_cursor(self, line, col):
        curses.setsyx(self.line + line, self.col + col)


@contextmanager
def winfocus(win):
    """Window context - makes the code more ergonomic and
    schedules an update upon exit"""
    try:
        yield win
    finally:
        win.noutrefresh()


class Windows:
    """The standard collection of windows we use, for ergonomy"""

    def __init__(self):
        self.title = None
        self.prompt = None
        self.binput = None
        self.menu = None


class UiCursesBase:
    """
    Base classe for the curses UI that provides the functionality but not the layout.

    Derived classes should def layout(self), override the prompt, etc.
    They can also def input_process to provide additional bindings.
    """

    prompt = "> "
    edit_actions: Dict[bytes, object] = {}
    edit_actions_default = {
        b"KEY_ENTER": View.key_enter,
        b"KEY_DOWN": View.key_down,
        b"KEY_UP": View.key_up,
        b"KEY_PPAGE": View.key_pgup,
        b"KEY_NPAGE": View.key_pgdown,
        b"KEY_HOME": View.key_home,
        b"KEY_END": View.key_end,
        b"KEY_BACKSPACE": View.key_backspace,
        b"KEY_DC": View.key_delete,
        b"KEY_LEFT": View.key_left,
        b"KEY_RIGHT": View.key_right,
        b"^A": View.key_bol,
        b"^E": View.key_eol,
        b"^U": View.key_killbol,
        b"^K": View.key_killeol,
        b"^[KEY_BACKSPACE": View.key_killwordleft,
    }

    def __init__(self):
        self.stdscr = None
        self.win = Windows()

    def start(self):
        """
        Based on
        https://github.com/enthought/Python-2.7.3/blob/master/Lib/curses/wrapper.py
        """
        # Initialize curses
        self.stdscr = curses.initscr()
        # Turn off echoing of keys, and enter cbreak mode,
        # where no buffering is performed on keyboard input
        curses.noecho()
        curses.cbreak()
        # In keypad mode, escape sequences for special keys
        # (like the cursor keys) will be interpreted and
        # a special value like curses.KEY_LEFT will be returned
        self.stdscr.keypad(True)
        # Start color, too.  Harmless if the terminal doesn't have
        # color; user can test with has_color() later on.  The try/catch
        # works around a minor bit of over-conscientiousness in the curses
        # module -- the error return from C start_color() is ignorable.
        try:
            curses.start_color()
            # Reset colors to default
            curses.use_default_colors()
        except curses.error:
            pass
        self.layout()

    def end(self):
        if not self.stdscr:
            return
        # Set everything back to normal
        if curses.has_colors():
            curses.use_default_colors()
        self.stdscr.keypad(0)
        curses.echo()
        curses.nocbreak()
        curses.endwin()

    def layout(self):
        raise NotImplementedError

    def max_items(self):
        return self.win.menu.height

    def show(self, view):
        # Update menu:
        with winfocus(self.win.menu) as win:
            win.erase()
            view.screen_height_set(self.max_items())
            for i, item in enumerate(view.screen_items()):
                win.addstr(i, 0, item)
            if view.screen_selected_line() is not None:
                win.addstr(
                    view.screen_selected_line(),
                    0,
                    view.selected_item(),
                    curses.A_REVERSE,
                )
        # Update input:
        with winfocus(self.win.input) as win:
            win.erase()
            win.addstr(0, 0, view.binput.string)
        # Update title, with status:
        with winfocus(self.win.title) as win:
            win.erase()
            win.addstr(0, 0, view.title)
            status = "%s/%d" % (
                str(
                    view.selected_idx + 1
                    if view.selected_idx is not None
                    else view.selected_idx
                ),
                len(view.items),
            )
            win.addstr(0, win.width - len(status) - 1, status)
        # Position cursor in prompt:
        self.win.prompt.set_cursor(0, len(self.prompt) + view.binput.pos)
        # Refresh screen:
        curses.doupdate()

    def input_read(self, nodelay):
        self.win.input.win.nodelay(nodelay)
        key = self.win.input.win.getch()
        if key == -1:
            return key, None
        if key in {curses.KEY_ENTER, 10, 13}:
            return curses.KEY_ENTER, b"KEY_ENTER"
        if key == 127:
            return curses.KEY_BACKSPACE, b"KEY_BACKSPACE"
        keyname = curses.keyname(key)
        if key == 27:  # ESC
            self.win.input.win.nodelay(True)
            k = self.win.input.win.getch()
            if k != -1:
                keyname += curses.keyname(k)
        logger.debug(f"key {key} name {keyname}")
        return key, keyname

    def interact(self, view):
        # Generate first item:
        view.item_generate()
        # Set nonblocking if we have more items to generate, otherwise
        # set blocking mode:
        key, keyname = self.input_read(bool(view.item_generator))
        return self.input_process(view, key, keyname)

    def input_process(self, view, key, keyname):
        if key == -1:
            return False
        # Check if it's a custom edit_action
        action = self.edit_actions.get(keyname)
        if action:
            return action(view, key, keyname)
        # Check if it's a default edit_action
        action = self.edit_actions_default.get(keyname)
        if action:
            return action(view, key, keyname)
        # Otherwise, check if it's a regular typed char:
        char = None
        try:
            char = chr(key)
        except ValueError:
            pass
        if char and char.isprintable():
            view.typed(char)
        return False


class UiCursesSimple(UiCursesBase):
    """
    Provides the following layout:

    title----------------------------
    prompt- input--------------------
    menu-----------------------------
    |||||||||||||||||||||||||||||||||
    |||||||||||||||||||||||||||||||||
    """

    def layout(self):
        lines = curses.LINES
        cols = curses.COLS
        self.win.title = CursesWin("title", 1, cols, 0, 0)
        with winfocus(CursesWin("prompt", 1, len(self.prompt) + 1, 1, 0)) as win:
            win.addstr(0, 0, self.prompt)
            self.win.prompt = win
        inputwidth = cols - len(self.prompt)
        with winfocus(CursesWin("input", 1, inputwidth, 1, len(self.prompt))) as win:
            win.win.keypad(True)
            self.win.input = win
        menulines = lines - 2
        self.win.menu = CursesWin("menu", menulines, cols, 2, 0)


@contextmanager
def context(uiclass=UiCursesSimple):
    ui = uiclass()
    try:
        ui.start()
        yield ui
    finally:
        ui.end()


def main():
    with context() as ui:
        items = [str(i) * 10 for i in range(0, ui.max_items())]
        ui.set("> ", items)
        ui.refresh()
        ui.getkey()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.

import setuptools

if __name__ == "__main__":
    setuptools.setup()



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.

import unittest

import tuzue.binput


class TestInput(unittest.TestCase):
    def test_backspace(self):
        binput = tuzue.binput.Binput()
        for c in "test":
            binput.typed(c)
        self.assertEqual(binput.string, "test")
        binput.key_backspace()
        self.assertEqual(binput.string, "tes")

    def test_arrows_backspace(self):
        binput = tuzue.binput.Binput("test")
        self.assertEqual(binput.string, "test")
        binput.key_left()
        binput.key_backspace()
        self.assertEqual(binput.string, "tet")
        binput.key_left()
        binput.key_left()
        binput.key_backspace()
        self.assertEqual(binput.string, "tet")
        binput.key_right()
        binput.key_backspace()
        self.assertEqual(binput.string, "et")

    def test_arrows_delete(self):
        binput = tuzue.binput.Binput("tst")
        self.assertEqual(binput.string, "tst")
        binput.key_delete()
        self.assertEqual(binput.string, "tst")
        binput.key_left()
        binput.key_delete()
        self.assertEqual(binput.string, "ts")
        binput.key_delete()
        self.assertEqual(binput.string, "ts")
        binput.key_left()
        binput.key_left()
        binput.key_delete()
        self.assertEqual(binput.string, "s")
        binput.key_delete()
        binput.key_delete()

    def test_boleol_delete(self):
        binput = tuzue.binput.Binput("tst")
        self.assertEqual(binput.string, "tst")
        binput.key_bol()
        self.assertEqual(binput.string, "tst")
        binput.key_delete()
        self.assertEqual(binput.string, "st")
        binput.key_eol()
        binput.key_backspace()
        self.assertEqual(binput.string, "s")

    def test_killbol(self):
        binput = tuzue.binput.Binput("asdf zxcv")
        for i in range(5):
            binput.key_left()
        binput.key_killbol()
        self.assertEqual(binput.string, " zxcv")

    def test_killeol(self):
        binput = tuzue.binput.Binput("asdf zxcv")
        for i in range(4):
            binput.key_left()
        binput.key_killeol()
        self.assertEqual(binput.string, "asdf ")

    def test_key_killwordleft(self):
        binput = tuzue.binput.Binput("asdf")
        binput.key_killwordleft()
        self.assertEqual(binput.string, "")
        self.assertEqual(binput.pos, 0)
        binput = tuzue.binput.Binput("asdf zxcv 1234")
        binput.key_killwordleft()
        self.assertEqual(binput.string, "asdf zxcv ")
        binput = tuzue.binput.Binput("asdf zxcv 1234")
        binput.pos = 0
        binput.key_killwordleft()
        self.assertEqual(binput.string, "asdf zxcv 1234")
        binput = tuzue.binput.Binput("asdf zxcv 1234")
        binput.pos = 9
        binput.key_killwordleft()
        self.assertEqual(binput.string, "asdf  1234")
        binput = tuzue.binput.Binput("asdf zxcv 1234")
        binput.pos = 10
        binput.key_killwordleft()
        self.assertEqual(binput.string, "asdf 1234")
        binput = tuzue.binput.Binput("asdf zxcv 1234")
        binput.pos = 8
        binput.key_killwordleft()
        self.assertEqual(binput.string, "asdf v 1234")



# Copyright (C) 2022 Leandro Lisboa Penz <lpenz@lpenz.org>
# This file is subject to the terms and conditions defined in
# file 'LICENSE', which is part of this source code package.
"""
A Binput object has the state of the input line
"""


class Binput:
    def __init__(self, string=""):
        self.string = string
        self.pos = len(string)

    def typed(self, char):
        self.string = self.string[: self.pos] + char + self.string[self.pos :]
        self.pos += 1

    def key_backspace(self):
        if self.string and self.pos > 0:
            self.string = self.string[: self.pos - 1] + self.string[self.pos :]
            self.pos -= 1

    def key_delete(self):
        if self.string and self.pos < len(self.string):
            self.string = self.string[: self.pos] + self.string[self.pos + 1 :]

    def key_left(self):
        if self.string and self.pos > 0:
            self.pos -= 1

    def key_right(self):
        if self.pos < len(self.string):
            self.pos += 1

    def key_bol(self):
        self.pos = 0

    def key_eol(self):
        self.pos = len(self.string)

    def key_killbol(self):
        self.string = self.string[self.pos :]
        self.pos = 0

    def key_killeol(self):
        self.string = self.string[: self.pos]

    def key_killwordleft(self):
        if self.pos == 0:
            return
        start = self.pos - 1
        while self.string[start].isspace() and start > 0:
            start -= 1
        while not self.string[start].isspace() and start > 0:
            start -= 1
        if start == 0:
            self.string = ""
            self.pos = 0
            return
        start += 1
        left = self.string[0:start]
        right = self.string[self.pos :]
        self.string = left + right
        self.pos = start
